/*****
 * xscreenshot.c
 * Take screenshot, with selectable area.
 * Copyright (C) James Budiono, 2014
 * License: GNU GPL Version 3 or later.
 *****/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X11/Xregion.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "pnglite.h"

#define DEFAULT_SCREENSHOT_PREFIX "xscreenshot"
#define DEFAULT_SCREENSHOT_TAKEN_MSG "Screenshot taken!"
#define MSG_FONTNAME "dejavu sans"
#define MSG_FONTSIZE "30"
#define MSG_DELAY 3

// so we don't pass thousands of parameters each time
typedef struct
{
	Display *display;
	Window	root_window;
	int     screen;
	int     width, height;
	unsigned int hotkey_keycode, hotkey_keycode2;
	
	GC     pen;	
	Cursor grab_cursor;
	BOX    anchor; //x1,y1=anchor, x2,y2=current pointer
	int    select_in_progress;
	int    mouse_grabbed;

	char *screenshot_path;
	int screenshot_index;
	
	int screenshot_delay;
} ProgState;


////////////// SCREEN CAPTURE ROUTINE //////////////
// get shift count based on mask, assumes LSB
int get_shift (int mask)
{
	int n=0, test=1;
	while (n < sizeof(mask)*8) {
		if (mask & test) return n;
		n++; test<<=1;
	}
	return 0;
}

// get 24-bit rgb data from XImage, ZPixmap format
// only supports direct/truecolor 15/16/24 bit, assumes LSB
unsigned char *extract_rgb24_from_ximage (XImage *image)
{
	unsigned char *out=NULL;
	int x,y,index;
	int width, height;
	unsigned long pix;
	
	int red_shift, green_shift, blue_shift;
	int red_mask, green_mask, blue_mask;
	double red_factor=1.0, green_factor=1.0, blue_factor=1.0;
	
	width       = image->width;
	height      = image->height;
	red_mask    = image->red_mask;
	green_mask  = image->green_mask;
	blue_mask   = image->blue_mask;
	red_shift   = get_shift (red_mask);
	green_shift = get_shift (green_mask);
	blue_shift  = get_shift (blue_mask);
	
	switch (image->depth) {
		case 15:
		case 16:
			red_factor   = 255.0/(red_mask   >> red_shift);
			green_factor = 255.0/(green_mask >> green_shift);
			blue_factor  = 255.0/(blue_mask  >> blue_shift);		
		case 24:
		case 32:
			break;
		default:
			fprintf (stderr, "Unsupported bit depth: %d.\n", image->depth);
			return NULL;
	}
	
	out = malloc (width * height * 3); // 24-bit RGB (3 bytes per pixel)	
	for (index=0, y=0; y < height; y++)
	{
		for (x=0; x < width; x++)
		{
			pix = XGetPixel (image, x,y);
			out[index++] = ((pix & red_mask)   >> red_shift) * red_factor;
			out[index++] = ((pix & green_mask) >> green_shift) * green_factor;
			out[index++] = ((pix & blue_mask)  >> blue_shift) * blue_factor;
		}
	}
	return out;
}

// take screenshot (if enabled), using pnglite
void take_screenshot (ProgState *st, BOX rect) // x2,y2 is width, height
{
	XImage *image;
	png_t png;
	unsigned char *buf;
	
	char *fname=NULL;
	struct stat sb;
	
	// sanity check - if either height or width is zero, abort
	if (!rect.x2 || !rect.y2) return;
	
	// find a file that does not exist yet
	while (1) {
		asprintf (&fname, "%s%.5d.png", st->screenshot_path, st->screenshot_index++);
		if (stat (fname, &sb)) break;
		else free (fname);
	};

	// do actual screenshot
	image = XGetImage (
		st->display, st->root_window, rect.x1, rect.y1, rect.x2, rect.y2, 
		AllPlanes, ZPixmap
	);
	buf = extract_rgb24_from_ximage (image);
	
	// save it
	if (buf) {
		png_open_file_write (&png, fname);
		png_set_data (&png, image->width, image->height, 8, PNG_TRUECOLOR, buf); //8 bits per channel.
		png_close_file (&png);
		free (buf);	
	}
	free (fname);
	XDestroyImage (image);
	fprintf (stderr, "screenshot taken\n");
}
///////////////// END OF SCREEN CAPTURE ROUTINE ///////////////


// enable notification when our hotkey is pressed
char *grab_hotkey (ProgState *st, char *hotkey)
{
	char *hotkey_name = hotkey;
	unsigned int keycode = 0, keycode2 = 0;
	KeySym keysym, keysym2;
	
	keysym = XStringToKeysym (hotkey);
	if (keysym != NoSymbol)
		keycode = XKeysymToKeycode (st->display, keysym);
	else {
		keycode = atoi(hotkey);
		if (keycode) {
			keysym2 = XKeycodeToKeysym(st->display, keycode, 0);
			keycode2 = XKeysymToKeycode (st->display, keysym2);
			hotkey_name = XKeysymToString (keysym2);
		}
	}
		
	if (keycode)
		XGrabKey (st->display, keycode, AnyModifier, st->root_window, 
		          False, GrabModeAsync, GrabModeAsync);
	if (keycode2 && keycode2 != keycode)
		XGrabKey (st->display, keycode2, AnyModifier, st->root_window, 
		          False, GrabModeAsync, GrabModeAsync);	
	if (!keycode && !keycode2) 
	{
		fprintf (stderr, "I don't understand %s, bailing out.\n", hotkey);
		exit(1);
	}
	st->hotkey_keycode = keycode;
	st->hotkey_keycode2 = keycode2;
	return hotkey_name;
}


// grap or ungrab the mouse pointer
int grab_ungrab_mouse (ProgState *st, int grab)
{
	if (grab)
	{
		//fprintf (stderr, "mouse grab\n");
		XGrabPointer (st->display, st->root_window, False, 
					  ButtonPressMask|ButtonReleaseMask|PointerMotionMask,
					  GrabModeAsync, GrabModeAsync, None, st->grab_cursor, CurrentTime);
		
	} else {
		//fprintf (stderr, "mouse release\n");
		XUngrabPointer (st->display, CurrentTime);
	}
	st->mouse_grabbed = grab;	
}


// draw a rectangle, x2,y2 is width, height
draw_rect (ProgState *st, BOX rect)
{
	//fprintf (stderr, "%d, %d, %d, %d\n", box->x1, box->y1, width, height);
	XDrawRectangle (
		st->display, st->root_window, st->pen, rect.x1, rect.y1, rect.x2, rect.y2
	);
}

// convert anchor points to a rect (x,y,width,height)
BOX convert_anchor_to_rect (BOX *anchor)
{
	BOX out;
	out.x1 = anchor->x1;
	if (anchor->x1 > anchor->x2) out.x1 = anchor->x2;
	out.y1 = anchor->y1;
	if (anchor->y1 > anchor->y2) out.y1 = anchor->y2;
	out.x2 = abs (anchor->x1 - anchor->x2);
	out.y2 = abs (anchor->y1 - anchor->y2);
	return out;
}

// convert anchor points to an area (x,y,width,height) 
// inclusive of the anchors themselves
BOX convert_anchor_to_area (BOX *anchor) {
	BOX out = convert_anchor_to_rect (anchor);
	out.x2++;
	out.y2++;
	return out;
}

// start selection
void start_selection (ProgState *st, int x, int y)
{
	st->anchor.x1 = st->anchor.x2 = x;
	st->anchor.y1 = st->anchor.y2 = y;
	st->select_in_progress = 1;
	draw_rect (st, convert_anchor_to_rect (&st->anchor));
}
// end active selection if currently in progress
void end_selection (ProgState *st)
{
	if (st->select_in_progress)
	{
		st->select_in_progress = 0;
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
	}
}
// draw selection box if selection is in progres
void draw_selection (ProgState *st, int x, int y)
{
	if (st->select_in_progress)
	{
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
		st->anchor.x2 = x;
		st->anchor.y2 = y;
		draw_rect (st, convert_anchor_to_rect (&st->anchor));
	}
}

//
void notify(ProgState *st, char *msg, int x, int y, int delay) {
	int msglen;
	
	if (!msg) return; 
	msglen = strlen(msg);
	
	XDrawString(st->display, st->root_window, st->pen, x, y, msg, msglen);
	XSync(st->display, False);
	sleep (delay);
	XDrawString(st->display, st->root_window, st->pen, x, y, msg, msglen);
	XSync(st->display, False);	
}

void delayed_capture(ProgState *st)
{
	BOX rect = { .x1=0, .y1=0, .x2=st->width, .y2=st->height };

	fprintf (stderr, "screenshot in %d seconds\n", st->screenshot_delay);
	sleep (st->screenshot_delay);
	take_screenshot (st, rect);
	fprintf (stderr, "screenshot taken\n");
	notify (st, "screenshot taken\n", 0, atoi(MSG_FONTSIZE), MSG_DELAY);
}


// main event loop
int event_loop (ProgState *st)
{
	XEvent ev;
	int running=1;
	int state;
	
	while (running)
	{
next_event:
		XNextEvent(st->display, &ev);
		switch(ev.type)
		{
			case ButtonPress:
				if (ev.xbutton.button == Button1)
					start_selection (st, ev.xbutton.x, ev.xbutton.y);
				break;
				
			case ButtonRelease:
				switch (ev.xbutton.button)
				{
					case Button1:
						if (st->select_in_progress)
						{	// selection done
							end_selection (st);
							take_screenshot (st, convert_anchor_to_area (&st->anchor));
							if (!st->hotkey_keycode) {
								grab_ungrab_mouse (st, 0);
								running=0;
							}
						}
						break;
						
					case Button3:
						if (st->select_in_progress)
						{
							end_selection (st);
						}
						else
						{	// leave if not running in multi shots mode
							if (!st->hotkey_keycode)
							{
								grab_ungrab_mouse (st, 0);
								running=0;
							}
						}
						
						break;
				}
				break;
				
			case MotionNotify:
				draw_selection (st, ev.xmotion.x, ev.xmotion.y);
				break;
			
			case KeyPress:
				if (ev.xkey.keycode == st->hotkey_keycode ||
				    ev.xkey.keycode == st->hotkey_keycode2 )
				{
					state=0;
					if (ev.xkey.state & ShiftMask) state ^= 1;
					if (ev.xkey.state & ControlMask) state ^= 2;
					if (ev.xkey.state & Mod1Mask) state ^= 4; // Alt
					switch (state)
					{
						case 0: // bare hotkey - enable/disable
							if (st->mouse_grabbed)
							{
								if (st->select_in_progress) 
									end_selection (st);
								grab_ungrab_mouse (st, 0);
							}
							else
							{
								grab_ungrab_mouse (st, 1);
							}
							break;
							
						case 1: // shift - delayed fullscreen shot
							delayed_capture (st);
							break;
							
						case 3: // ctrl-shift - exit
							if (st->select_in_progress) 
								end_selection (st);
							if (st->mouse_grabbed)
								grab_ungrab_mouse (st, 0);
							running=0;
							break;
					}
				}
		}
	}
	return 0;
}


// ***************** Entry point *************
int main (int argc, char*argv[])
{
	ProgState st;
	char *hotkey = NULL;
	XGCValues values;
	unsigned long gcmask;
	XFontStruct *font;
	int opt;

	// default values
	st.hotkey_keycode = st.hotkey_keycode2 = 0;
	st.screenshot_index = 0;
	asprintf (&st.screenshot_path, "%s/" DEFAULT_SCREENSHOT_PREFIX, getenv("HOME"));
	st.screenshot_delay = 5; // default delay is 5 seconds
	
	// argument processing	
	while ((opt = getopt(argc, argv, "n:p:k:d:h")) != -1) 
	{
		switch (opt) 
		{
			case 'k':
				if (optarg) hotkey = optarg;
				break;
			case 'p':
				if (optarg) st.screenshot_path = optarg;
				break;
			case 'n':
				st.screenshot_index = atoi (optarg);
				break;
			case 'd':
				st.screenshot_delay = atoi (optarg);
				break;
			case 'h':
				fprintf (stderr,
"Usage: xscreenshot [-k hotkey] [-p path] [-n index]\n"
"Copyright (C) James Budiono, 2014. License: GPL version 3.0 or later\n\n"
" -h      : this text\n"
" -k key  : multiple shots mode, use 'key' as hotkey (default none)\n"
" -p path : path/filename for screenshot (default is $HOME/" DEFAULT_SCREENSHOT_PREFIX ")\n"
" -n index: the starting index used for screenshot filename\n"
" -d delay: the delay (in seconds) for fullscreen screenshot (default 5)\n\n"
"If you don't use -k, the program will take one screenshot and exit.\n"
"If you do: press hotkey to enter screenshot taking mode, take as many\n"
"  screenshots, and press hotkey to return back to normal mode.\n"
"  Press Ctrl-Shift-hotkey to terminate program\n"
);
				return 0;
		}
	}

	fprintf (stderr,
"--- Mouse usage ---\n"
"Left-click  - press to start, move to select, release to take shot\n"
"Right-click - cancel current selection, (in single-shot: once more to exit.)\n");

	// initialisations
	png_init (0,0);
    st.display = XOpenDisplay(NULL);
    if (st.display == NULL) {
        fprintf(stderr, "Cannot open display\n");
        exit(1);
    }
    st.screen      = DefaultScreen (st.display);
    st.root_window = DefaultRootWindow (st.display);
    st.width       = DisplayWidth(st.display, st.screen);
    st.height      = DisplayHeight(st.display, st.screen);
    st.grab_cursor = XCreateFontCursor (st.display, XC_crosshair);

	values.function = GXinvert;
	values.subwindow_mode = IncludeInferiors;
	values.line_width = 1;
	gcmask = GCFunction | GCSubwindowMode | GCLineWidth;
	font = XLoadQueryFont(st.display, "-*-" MSG_FONTNAME "-*-*-*--" MSG_FONTSIZE "-*-*-*-*-*-*-*");
	if (font)
	{
		values.font = font->fid;
		gcmask |= GCFont;
	}
    st.pen = XCreateGC (st.display, st.root_window, gcmask, &values);
    
    if (hotkey) {
		char *hotkey_name = grab_hotkey (&st, hotkey);
		fprintf (stderr,
"--- Multiple shots mode enabled ---\n"
"%s            - to toggle screenshot taking mode\n"
"Shift-%s      - fullscreen delayed screenshot\n"
"Ctrl-Shift-%s - exit\n", hotkey_name, hotkey_name, hotkey_name);
	}
    else grab_ungrab_mouse (&st, 1);
    fprintf(stderr, "---\n");

	// run
	st.select_in_progress = 0;
    event_loop (&st);
    
    // cleanup
    XFreeGC (st.display, st.pen);
	XFlush (st.display);
	XCloseDisplay (st.display);
	return 0;
}

